/*
   Copyright (C) 1998,1999 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpsim; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


// !!!!!!!!!!!!!!!! NOTE !!!!!!!!!!!!!!!!!
//
// This file is not used in version 0.0.14
// I'm postponing CORBA ...
#if 0

/*
  cli-corba.cc  - This provides a corba (common object request broker
     architecture) wrapper around gpsim's cli (command line interface).

*/
#include <stdio.h>
#include <stdlib.h>
#include <orb/orbit.h>
#include <pthread.h>
#include "../src/gpsim_def.h"
#include "../idl/gpsim-corba.h"

#include "input.h"

// External cli stuff that should really go into an h-file.
extern void welcome(void);
void init_parser(void);
int yyparse(void);
extern int yydebug;
extern int quit_parse;


#ifdef HAVE_GUI
  void quit_gui(void);
#endif

// CORBA stuff:
extern Gpsim_Register Gpsim_Register_client;
static Gpsim_Register bec;


// pthread stuff. Two threads are used to start the cli. The first
// thread controls the readline library loop while the second ensures
// corba is ready before the cli begins.

static  pthread_t th_a, th_b;
static  void * retval;

// Here's a simple mutex (MUTual EXclusion) structure used to 
// synchronize the cli and corba.

struct synchronize {
  pthread_mutex_t lock;
  pthread_cond_t notORBitting;
  pthread_cond_t notWaiting;
};


//-----------------------------------------------------------
//
// init_threads - Initialize the synchonize mutex
//

void init_threads(struct synchronize * b)
{
  pthread_mutex_init(&b->lock, NULL);
  pthread_cond_init(&b->notORBitting, NULL);
  pthread_cond_init(&b->notWaiting, NULL);
}

//-----------------------------------------------------------
//
// wait_for_corba - suspend the thread until corba is ready
// 
//

void wait_for_corba(struct synchronize * b)
{
  pthread_mutex_lock(&b->lock);

  printf("waiting for corba. pid %d\n",getpid());

  pthread_cond_wait(&b->notORBitting, &b->lock);

  printf("back from wait_for_corba .. cond_wait. pid %d\n",getpid());

  pthread_cond_signal(&b->notWaiting);

  pthread_mutex_unlock(&b->lock);

}

//-----------------------------------------------------------
//
// corba_is_ready - 
//

void corba_is_ready(struct synchronize * b)
{

  pthread_mutex_lock(&b->lock);

  pthread_cond_signal(&b->notORBitting);

  printf("corba should be ready. pid %d\n",getpid());

  pthread_mutex_unlock(&b->lock);

}


struct synchronize sync_thread_struc;


//-----------------------------------------------------------
//
//  cli_thread - This thread handles gpsim's cli. This code used
//     to be that little snippet at the end of gpsim's main. Now
//     that gpsim has to handle more than one user interface, this
//     occupies a thread. 
//

void * cli_thread(char *d)
{
  int i;

  // Do not process the cli until corba is up and running.
  wait_for_corba(&sync_thread_struc);

  // splash screen for the cli:
  welcome();

  // Loop until the user has had enough fun. 

  do {
    init_parser();
    i = yyparse();

  } while(!quit_parse);


  // When the user is done then so is gpsim. Without this exit,
  // only this thread will terminate - the other threads would 
  // still continue.

#ifdef HAVE_GUI
  quit_gui();
#endif

  exit(0);

  // never happens
  return NULL;
}


//-----------------------------------------------------------
//
// consumer - this thread does nothing other than signaling 
//    that corba is ready. How does it do that? Well, what
//    happens is that this thread is created before corba is
//    initialized, but doesn't start executing until after
//    corba (orbit) relinquishes control of the cpu. ORBit
//    will relinquish control sometime after CORBA_ORB_run 
//    is called. Fortuitous timing? Perhaps.
//

void * consumer(void * data)
{

  printf("in consumer thread. pid %d\n",getpid());

  corba_is_ready(&sync_thread_struc);

  printf("consumer just made corba ready. pid %d\n",getpid());

  return NULL;
}

/*
static CORBA_Object
do_echoString(PortableServer_Servant servant,
	      CORBA_char *astring,
	      CORBA_long *outnum,
	      CORBA_Environment *ev)
{
  g_message("[cli server] %s. pid %d", astring, getpid());

  *outnum = rand() % 100;

  return CORBA_Object_duplicate(echo_client, ev);
}
*/


//-----------------------------------------------------------
//
// cli_init - initialize the command line interface.
//
// The cli runs within a thread. This routine initializes the
// thread interface
//

int
cli_corba_init ( CORBA_Environment *ev)

{

    //
    // Initialize the threads to spawn the cli
    //

    printf("in cli_init. pid %d\n",getpid());
    
    init_threads(&sync_thread_struc);

    pthread_create(&th_a, NULL, cli_thread, 0);
    pthread_create(&th_b, NULL, consumer, 0);

    printf("created th_a. pid %d\n",getpid());

    return 0;
}


//-----------------------------------------------------------
//
// Experimental corba code
//

#if 0

int
cli_corba_init ( CORBA_Environment *ev)

{

    CORBA_long rv;

    char buf[30];
    int i;

    int niters = 3;


    //
    // Initialize the threads to spawn the cli
    //

    printf("in cli_corba_init. pid %d\n",getpid());
    
    init_threads(&sync_thread_struc);

    pthread_create(&th_a, NULL, cli_thread, 0);
    pthread_create(&th_b, NULL, consumer, 0);

    printf("created th_a. pid %d\n",getpid());


    printf("cli corba\n");

    for(i = 0; i < niters; i++) {
	    g_snprintf(buf, sizeof(buf), "cli: Hello, world [%d]", i);
	    bec = Gpsim_Register_echoString(Gpsim_Register_client, buf, &rv, ev);
	    if(ev->_major != CORBA_NO_EXCEPTION) {
	      printf("we got exception %d from echoString!\n", ev->_major);
	      return 1;
	    }
	    printf("cli- looping\n");



	    CORBA_Object_release(Gpsim_Register_client, ev);
	    if(ev->_major != CORBA_NO_EXCEPTION) {
	      printf("we got exception %d from release!\n", ev->_major);
	      return 1;
	    }
	    g_message("[client] %d", rv);

	    Gpsim_Register_client = bec; bec = CORBA_OBJECT_NIL;
    }


    return 0;
}

#endif
#endif
