/*
   Copyright (C) 1999 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpsim; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <iostream.h>
#include <iomanip.h>
#include <string>

#include "command.h"
#include "cmd_processor.h"

#include "../src/pic-processor.h"

cmd_processor c_processor;

static cmd_options cmd_processor_options[] =
{
  "list",1,    OPT_TT_BITFLAG,
  NULL,0,0
};


cmd_processor::cmd_processor(void)
{ 
  name = "processor";

  brief_doc = string("Select & Display processors");

  long_doc = string ("processor [new_processor_type [new_processor_name]] | [list]\
\n\tIf no new processor is specified, then the currently defined processor(s)\
\n\twill be displayed. To see a list of the processors supported by gpsim,\
\n\ttype 'processor list'.  To define a new processor, specify the processor\
\n\ttype and name.\
\n\n\texamples:
\n\n\tprocessor               // Display the processors you've already defined.\
\n\tprocessor list          // Display the list of processors supported.\
\n\tprocessor p16cr84 fred  // Create a new processor.\
\n\tprocessor p16c74 wilma  // and another.\
\n\tprocessor p16c65        // Create one with no name.\
\n");

  op = cmd_processor_options; 
}


void cmd_processor::processor(void)
{

  dump_processor_list();

}

void cmd_processor::processor(int bit_flag)
{

  switch(bit_flag)
    {

    case 1:
      display_available_processors();
      break;

    }

}


void cmd_processor::processor(char * processor_type, char * processor_new_name)
{

  new_processor(add_processor( processor_type,  processor_new_name));

}
