

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <gtk/gtk.h>

#include <gdk/gdktypes.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

#include <iostream.h>

#include "gtksheet/gtkcombobox.h"
#include "gtksheet/gtkbordercombo.h"
#include "gtksheet/gtkcolorcombo.h"
#include "gtksheet/gtksheet.h"
#include "gtksheet/gtksheetentry.h"

#include "../src/gpsim_def.h"
#include "../src/pic-processor.h"

#include "gui.h"

/*
 * --- Function prototypes
 */
void redisplay_prompt(void);

void gui_init(void);
void create_dispatcher (void);
void RegWindow_new_processor(Register_Window *rw, GUI_Processor *gp);
void StatusBar_new_processor(StatusBar_Window *sbw, GUI_Processor *gp);
void SourceBrowserAsm_new_processor(SourceBrowserAsm_Window *sbaw, GUI_Processor *gp);
void SourceBrowserOpcode_new_processor(SourceBrowserOpcode_Window *sbow, GUI_Processor *gp);
void SourceBrowserAsm_new_source(SourceBrowserAsm_Window *sbaw, GUI_Processor *gp);
void SourceBrowserOpcode_new_program(SourceBrowserOpcode_Window *sbow, GUI_Processor *gp);
GUI_Processor *new_GUI_Processor(void);

void init_link_to_gpsim(GUI_Processor *gp);
void link_src_to_gpsim(GUI_Processor *gp);
GtkWidget *CreateRegisterWindow(GUI_Processor *gp);
GtkWidget *CreateSourceBrowserOpcodeWindow(GUI_Processor *gp);
GtkWidget *CreateSourceBrowserAsmWindow(GUI_Processor *gp);

extern StatusBar_Window *status_bar_window;
extern Register_Window *register_window;
/* 
 * --- Global variables
 */

GUI_Processor *gp=NULL;

GSList *gui_processors=NULL;


/*------------------------------------------------------------------
 * quit_gui
 *
 */

void quit_gui(void)
{
    gtk_main_quit();
}

/*------------------------------------------------------------------
 * gui_new_processor - Add a new processor to the register viewer
 *
 * This routine adds another pic processor to the list of currently
 * simulated processors (as of 0.0.14 though, you're still limited
 * to a list of one). It then notifies each child window. Finally
 * a communication link between the gui and the simulator is established.
 * (This was a corba link, but now it consists of direct calls...)
 */

void gui_new_processor (pic_processor *p)
{

  // printf("gui is adding a new processor\n");

  // Create an gui representation of the new processor

  gp->p = p;
  p->gp = gp;

  // Add it to the list 

  gui_processors = g_slist_append(gui_processors,gp);

  RegWindow_new_processor(register_window, gp);
  StatusBar_new_processor(status_bar_window, gp);
  SourceBrowserOpcode_new_processor((SourceBrowserOpcode_Window*)gp->program_memory, gp);
  SourceBrowserAsm_new_processor((SourceBrowserAsm_Window*)gp->source_browser, gp);

  init_link_to_gpsim(gp);
  //  redisplay_prompt();

}

/*------------------------------------------------------------------
 *
 */
void gui_new_program (pic_processor *p)
{

  // FIX ME - need to search for *p in the gp list...
  if(gp)
    {
      SourceBrowserOpcode_new_program((SourceBrowserOpcode_Window*)gp->program_memory, gp);
      link_src_to_gpsim( gp);
      //      redisplay_prompt();
    }
}

/*------------------------------------------------------------------
 *
 */
void gui_new_source (pic_processor *p)
{

  // FIX ME - need to search for *p in the gp list...
  if(gp)
    {
      SourceBrowserOpcode_new_program((SourceBrowserOpcode_Window*)gp->program_memory, gp);
      SourceBrowserAsm_new_source((SourceBrowserAsm_Window*)gp->source_browser, gp);
      link_src_to_gpsim( gp);
      //      redisplay_prompt();

    }
}

/*------------------------------------------------------------------
 * update_program_memory
 *
 */
void update_program_memory(GUI_Processor *gp, unsigned int reg_number)
{

  //  printf("program memory needs to be updated\n");

}
/*------------------------------------------------------------------
 * gui_update_object
 *
 * Each 'thing' that the gui displays about a simulated pic has an
 * associated cross reference structure. Sometimes these 'things' 
 * displayed in more than one place (like the status register).
 * Each graphical instance has its own structure. All of the structures
 * pertaining to the same pic object (again, like the status register)
 * are stored in a singly-linked list. This routine scans through
 * this list and updates each instance of the object.
 */

void gui_update_object(gpointer gui_xref,int new_value)
{
  GSList *gp;
  cross_reference_to_gui *xref;

  if(gui_xref)
    for(gp= (GSList *)gui_xref; gp; gp = gp->next)
      {
	if(gp->data)
	  {
	    xref = (cross_reference_to_gui *)gp->data;
	    xref->update(xref,new_value);
	  }
      }


}


/*------------------------------------------------------------------
 * gui_init
 *
 */

int gui_init (int argc, char **argv)
{
    FILE *fp;

  gtk_init (&argc, &argv);

  //  gp = (GUI_Processor *)malloc(sizeof(GUI_Processor));
  //  gp->windows = g_list_alloc();

  fp = fopen("~/.gpsim_config","r+");
  
  gp = new_GUI_Processor();

  gui_init();
  
  create_dispatcher();
  CreateRegisterWindow(gp);
  CreateSourceBrowserOpcodeWindow(gp);
  CreateSourceBrowserAsmWindow(gp);

//  config_retreive(CONFIG_DISPATCHER, gp->

  return(0);
}

void gui_main(void)
{
  
  redisplay_prompt();
  gtk_main();

}
