#include <stdio.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <string.h>
#include <iostream.h>

#include "gtksheet/gtkcombobox.h"
#include "gtksheet/gtkbordercombo.h"
#include "gtksheet/gtkcolorcombo.h"
#include "gtksheet/gtksheet.h"
#include "gtksheet/gtksheetentry.h"

#include "../src/gpsim_def.h"
#include "../src/pic-processor.h"
#include "gui.h"

#include <assert.h>

int SourceBrowser_change_view (struct _gui_object *_this, int view_state);

void SourceBrowserOpcode_update_line( SourceBrowserOpcode_Window *sbow, int address, int row);
void SourceBrowserAsm_update_line( SourceBrowserAsm_Window *sbaw, int address);
void SourceBrowserAsm_set_pc(SourceBrowserAsm_Window *sbaw, int address);


gint
key_press(GtkWidget *widget,
	  GdkEventKey *key, 
	  gpointer data)
{

  SourceBrowser_Window *sbw = (SourceBrowser_Window *) data;

  if(!sbw) return(FALSE);
  if(!sbw->gui_obj.gp) return(FALSE);
  if(!sbw->gui_obj.gp->p) return(FALSE);


  switch(key->keyval) {

  case 's':  // Single Step
  case 'S':
  case GDK_F7:
    sbw->gui_obj.gp->p->step(1);
    break;

  case 'o':  // Step Over Next instruction
  case 'O':
  case GDK_F8:
    sbw->gui_obj.gp->p->step_over();
    break;

  }

  return TRUE;
} 

static int
SourceBrowser_close(GtkWidget *widget,
		    SourceBrowser_Window *sbw)
{
  printf("closing browser\n");
  SourceBrowser_change_view((GUI_Object *)sbw,0);

  //gtk_widget_hide(widget);
  printf("browser should be closed\n");

  return(TRUE);

}

static int delete_event(GtkWidget *widget,
			GdkEvent  *event,
                        SourceBrowser_Window *sbw)
{
    SourceBrowser_change_view((GUI_Object *)sbw,0);
    return TRUE;
}

gint
button_press(GtkWidget *widget,
	     GdkEvent  *event, 
	     gpointer data)
{

    SourceBrowser_Window *sbw = (SourceBrowser_Window *) data;
    SourceBrowserOpcode_Window *sbow = (SourceBrowserOpcode_Window*)sbw;
  int break_row;

  if(!sbw) return(FALSE);
  if(!sbw->gui_obj.gp) return(FALSE);
  if(!sbw->gui_obj.gp->p) return(FALSE);


  if ( ((event->type == GDK_BUTTON_PRESS) &&
        (event->button.button == 3))
       ||
       ((event->type == GDK_2BUTTON_PRESS) &&
        (event->button.button == 1)) )

  {
      switch(sbw->wt) {

      case WT_asm_source_window:
/*	  index = gtk_editable_get_position(GTK_EDITABLE(sbaw->source_text));
	  e = gui_index_to_entry(index);
	  sbw->gui_obj.gp->p->toggle_break_at_line(0, e->line);*/
	  return FALSE;

      case WT_opcode_source_window:
	  break_row =  GTK_CLIST (sbow->clist)->focus_row;

	  if(!sbow->clist_rows)
	      return TRUE;      // no code is in this window
	  sbow->sbw.gui_obj.gp->p->toggle_break_at_address(break_row);
	  printf("Toggle at line %d\n",break_row);
	  return TRUE;

      case WT_list_source_window:
	  sbw->gui_obj.gp->p->toggle_break_at_line(1, break_row);
	  return TRUE;

      default:
	  printf("bad window type %d\n",sbw->wt);
      }

    }

  return FALSE;

} 

void SourceBrowser_update_line(cross_reference_to_gui *xref, int new_value)
{
  GUI_Processor *gp;
  SourceBrowser_Window *sbw;
  int address;

  gp = (GUI_Processor *)xref->parent_window;
  if(!gp) { printf("gp == null\n"); return;}

  assert(xref && xref->data);

  address = *(int *)xref->data;

  if(gp->source_browser)
      SourceBrowserAsm_update_line( (SourceBrowserAsm_Window*)gp->source_browser,  address);

  if(gp->program_memory)
      SourceBrowserOpcode_update_line( (SourceBrowserOpcode_Window*)gp->program_memory,  address, address);

}

void SourceBrowser_update(SourceBrowser_Window *sbw)
{
  char buffer[20];
  gint new_row,last_address, new_address;
  SourceBrowserOpcode_Window *sbow = (SourceBrowserOpcode_Window*)sbw;
  SourceBrowserAsm_Window *sbaw = (SourceBrowserAsm_Window*)sbw;

  
  pic_processor *pic;

  if(!sbw) return;     // this shouldn't happen

  // get a manageable pointer 
  pic = sbw->gui_obj.gp->p;

  //update the displayed values
  switch(sbw->wt) {

  case WT_asm_source_window:
//      new_row =  pic->program_memory[pic->pc.value]->get_src_line();
      //      new_address = pic->find_closest_address_to_line(0,new_row);
      new_address=pic->pc.value;
      SourceBrowserAsm_set_pc( sbaw, new_address);
      break;

  case WT_opcode_source_window:
      last_address = sbow->current_row;
      new_row = pic->pc.value;
      if(new_row != sbow->current_row)
      {

	  gtk_clist_freeze (GTK_CLIST (sbow->clist));

	  SourceBrowserOpcode_update_line( sbow, last_address, sbow->current_row);
	  //gtk_clist_set_row_style (GTK_CLIST (sbw->clist), sbw->current_row, normal_style);
	  sbow->current_row = new_row;
	  gtk_clist_set_row_style (GTK_CLIST (sbow->clist), sbow->current_row, current_line_number_style);


	  gtk_clist_thaw (GTK_CLIST (sbow->clist));
      }
      if(GTK_VISIBILITY_FULL != gtk_clist_row_is_visible (GTK_CLIST (sbow->clist),
							  sbow->current_row))
      {
	  gtk_clist_moveto (GTK_CLIST (sbow->clist), sbow->current_row, 0, .5, 0.0);
      }
      break;

  case WT_list_source_window:
      assert(0);/*
    last_address = pic->find_closest_address_to_line(1,sbw->current_row);
    new_row =  pic->program_memory[pic->pc.value]->get_lst_line();
    break;        */


  default:
    printf("SourceBrowser_update: bad window type %d\n",sbw->wt);
    return;
  }
  


}

GtkWidget *CreateSBW(SourceBrowser_Window *sbw)
{


  gint i;


  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *hbox;
  static GtkWidget *clist;
  GtkWidget *button;
  GtkWidget *separator;
  GtkWidget *scrolled_win;
  GtkWidget *check;

  GtkWidget *undo_button;
  GtkWidget *label;

  SourceBrowserOpcode_Window *sbow = (SourceBrowserOpcode_Window*)sbw;

  if(sbw->wt == WT_opcode_source_window)
  // FIXME opcode?
      sbow->clist_rows = 0;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  
  gtk_window_set_default_size(GTK_WINDOW(window),600,300);

  sbw->gui_obj.window = window;

  gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		      GTK_SIGNAL_FUNC(delete_event),
		      (gpointer) sbw);

  /* Add a signal handler for key press events. This will capture
   * key commands for single stepping, running, etc.
   */
  gtk_signal_connect(GTK_OBJECT(window),"key_press_event",
		     (GtkSignalFunc) key_press,
		     (gpointer) sbw);


  /* Add a signal handler for button press events. This will capture
   * commands for setting and/or clearing break points
   */
  gtk_signal_connect(GTK_OBJECT(window),"button_press_event",
		     (GtkSignalFunc) button_press,
		     (gpointer) sbw);

  
  gtk_window_set_title (GTK_WINDOW (window), sbw->gui_obj.name);

  gtk_container_set_border_width (GTK_CONTAINER (window), 0);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  // FIXME opcode?
  sbw->vbox=vbox;

}

int SourceBrowser_change_view (struct _gui_object *_this, int view_state)
{
  int depth; // not used

  switch(view_state)
  {
  case 0: // hide window?
      break;
  case 1: // show window?
      break;
  case 2: // toggle window?
      break;
  }

  if( (view_state==2||view_state==1) &&
      !GTK_WIDGET_VISIBLE(GTK_WIDGET(_this->window)))
    {

      gtk_window_set_default_size(GTK_WINDOW(_this->window), _this->width, _this->height);
      gtk_widget_set_uposition (_this->window, _this->x, _this->y);

      gtk_widget_show(_this->window);

    }
  else if (GTK_WIDGET_VISIBLE(GTK_WIDGET(_this->window)))
    {

      gdk_window_get_size(_this->window->window, &_this->width, &_this->height);

      gdk_window_get_root_origin (_this->window->window, 
			       &_this->x,
			       &_this->y);

      gtk_widget_hide(_this->window);
    }

}
