#include <stdio.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <string.h>
#include <iostream.h>

#include "gtksheet/gtkcombobox.h"
#include "gtksheet/gtkbordercombo.h"
#include "gtksheet/gtkcolorcombo.h"
#include "gtksheet/gtksheet.h"
#include "gtksheet/gtksheetentry.h"

#include "../src/gpsim_def.h"
#include "../src/pic-processor.h"
#include "gui.h"

#include <assert.h>

#define PROGRAM_MEMORY_WINDOW_COLUMNS 4   //yuk
#define DEFAULT_ROWS  256

#define PROFILE_COLUMN  0
#define ADDRESS_COLUMN  1
#define OPCODE_COLUMN   2
#define MNEMONIC_COLUMN 3

GtkWidget *CreateSBW(SourceBrowser_Window *sbw);
char *default_row_text[PROGRAM_MEMORY_WINDOW_COLUMNS];
extern int SourceBrowser_change_view (struct _gui_object *_this, int view_state);

static gint pc_row = 0;

static void 
clist_click_column (GtkCList *clist, gint column, gpointer data)
{

}

static void filter(char *clean, char *dirty, int max)
{

  int i=0,j;

  do {


    if(*dirty == '\t')

      for(j=0,*dirty++; j<8 && i%8; j++,i++)
	*clean++ = ' ';

    else if (*dirty <' ') 

      *dirty++;

    else

      *clean++ = *dirty++;



  }while(*dirty && ++i < max);

  *clean = 0;

}

void SourceBrowserOpcode_update_line( SourceBrowserOpcode_Window *sbow, int address, int row)
{
  if(!sbow) return;

  assert(sbow->sbw.wt == WT_opcode_source_window);

  if(address >= 0)
    {
      if(sbow->sbw.gui_obj.gp->p->program_memory[address]->isa() == instruction::BREAKPOINT_INSTRUCTION)
	gtk_clist_set_row_style (GTK_CLIST (sbow->clist), row, breakpoint_line_number_style);
      else
	  gtk_clist_set_row_style (GTK_CLIST (sbow->clist), row, normal_style);

    }

}

void SourceBrowserOpcode_new_program(SourceBrowserOpcode_Window *sbow, GUI_Processor *gp)
{
  char text_buffer[128];
  char clean_buffer[128];

  char *default_row_text[PROGRAM_MEMORY_WINDOW_COLUMNS];

  gint i;

  pic_processor *pic;

  if(sbow == NULL || gp == NULL)
      return;
  
  assert(sbow->sbw.wt==WT_opcode_source_window);

  sbow->sbw.gui_obj.gp = gp;
  gp->program_memory = (SourceBrowser_Window*)sbow;

  // get a manageable pointer to the processor
  pic = sbow->sbw.gui_obj.gp->p;

  

  gtk_clist_freeze (GTK_CLIST (sbow->clist));


  for(i=0; i < pic->program_memory_size(); i++)
    {
      sprintf(text_buffer, "0x%04X", pic->program_memory[i]->opcode);
      gtk_clist_set_text (GTK_CLIST (sbow->clist), i, OPCODE_COLUMN, text_buffer);

      filter(clean_buffer, pic->program_memory[i]->name(text_buffer), 128);

      gtk_clist_set_text (GTK_CLIST (sbow->clist),
			  i, MNEMONIC_COLUMN, 
			  clean_buffer);
    }

  sbow->current_row = pic->pc.value;
  gtk_clist_set_row_style (GTK_CLIST (sbow->clist), 0, current_line_number_style);

  gtk_clist_thaw (GTK_CLIST (sbow->clist));

}

void SourceBrowserOpcode_new_processor(SourceBrowserOpcode_Window *sbow, GUI_Processor *gp)
{
  char text_buffer[128];
  char *default_row_text[PROGRAM_MEMORY_WINDOW_COLUMNS];

  gint i;

  pic_processor *pic;

  if(sbow == NULL || gp == NULL)
    return;

  assert(sbow->sbw.wt==WT_opcode_source_window);
  
  // get a manageable pointer to the processor
  pic = sbow->sbw.gui_obj.gp->p;

  if(pic->program_memory_size() > sbow->clist_rows)
    {
      for(i=0; i < PROGRAM_MEMORY_WINDOW_COLUMNS; i++)
	default_row_text[i] = NULL;
      default_row_text[ADDRESS_COLUMN] = text_buffer;

      gtk_clist_freeze (GTK_CLIST (sbow->clist));

      for(i=sbow->clist_rows; i < pic->program_memory_size(); i++)
	{
	  sprintf (default_row_text[ADDRESS_COLUMN], "0x%04X", sbow->clist_rows++);
	  gtk_clist_append (GTK_CLIST (sbow->clist), default_row_text);
	  gtk_clist_set_row_style (GTK_CLIST (sbow->clist), i, normal_style);
	}

      for(i=0; i < pic->program_memory_size(); i++)
	{
	  sprintf(text_buffer, "0x%04X", pic->program_memory[i]->opcode);
	  gtk_clist_set_text (GTK_CLIST (sbow->clist), i, OPCODE_COLUMN, text_buffer);
	}

    }


  gtk_clist_thaw (GTK_CLIST (sbow->clist));


}


GtkWidget *CreateSourceBrowserOpcodeWindow(GUI_Processor *gp)
{

  static char *titles[] =
  {
    "profile", "address", "opcode", "instruction"
  };

  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *hbox;
  static GtkWidget *clist;
  GtkWidget *button;
  GtkWidget *separator;
  GtkWidget *scrolled_win;
  GtkWidget *check;

  GtkWidget *undo_button;
  GtkWidget *label;

  gint i;
  char address[16];
  static SourceBrowserOpcode_Window *sbow;

//  SourceBrowserOpcode_Data *sbow;

  sbow = (SourceBrowserOpcode_Window *) malloc(sizeof(SourceBrowserOpcode_Window));
  sbow->sbw.gui_obj.gp = NULL;
  sbow->sbw.gui_obj.window = NULL;
  sbow->sbw.wt = WT_opcode_source_window;

  sbow->column_titles = titles;
  sbow->columns = 4;

  sbow->sbw.gui_obj.gp = gp;
  gp->program_memory = (SourceBrowser_Window*)sbow;
  sbow->sbw.gui_obj.name = ".obj Browser";
  sbow->sbw.gui_obj.wc = WC_source;

  gp_add_window_to_list(gp, (GUI_Object *)sbow);
  sbow->sbw.gui_obj.change_view = SourceBrowser_change_view;


  CreateSBW((SourceBrowser_Window*)sbow);



  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_set_border_width (GTK_CONTAINER (scrolled_win), 5);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				  GTK_POLICY_AUTOMATIC, 
				  GTK_POLICY_AUTOMATIC);

  /* create GtkCList here so we have a pointer to throw at the 
   * button callbacks -- more is done with it later */
  clist = gtk_clist_new_with_titles (sbow->columns, sbow->column_titles);
  sbow->clist = clist;
  gtk_container_add (GTK_CONTAINER (scrolled_win), clist);
  gtk_signal_connect (GTK_OBJECT (clist), "click_column",
		      (GtkSignalFunc) clist_click_column, NULL);

  gtk_box_pack_start (GTK_BOX (sbow->sbw.vbox), scrolled_win, TRUE, TRUE, 0);
  gtk_clist_set_row_height (GTK_CLIST (clist), 18);
  gtk_widget_set_usize (clist, 300, 100);
  gtk_clist_set_selection_mode (GTK_CLIST (clist), GTK_SELECTION_EXTENDED);

  for (i = 0; i < sbow->columns; i++)
    {
      gtk_clist_set_column_width (GTK_CLIST (clist), i, 80);

      gtk_clist_set_column_auto_resize (GTK_CLIST (clist), i, FALSE);
      gtk_clist_set_column_justification (GTK_CLIST (clist), i,
					  GTK_JUSTIFY_LEFT);
      // %%% FIX ME
      //   Hide the profile column for now...
      if(i == 0)
	gtk_clist_set_column_visibility (GTK_CLIST (clist), i, FALSE);
    }


  for (i = 0; i < sbow->columns; i++)
    default_row_text[i] = NULL;


  default_row_text[ADDRESS_COLUMN] = address;

  for (i = 0; i < DEFAULT_ROWS; i++)
    {
      sprintf (default_row_text[ADDRESS_COLUMN], "0x%04X", sbow->clist_rows++);
      gtk_clist_append (GTK_CLIST (sbow->clist), default_row_text);
      gtk_clist_set_row_style (GTK_CLIST (sbow->clist), i, normal_style);

    }
  
  if (!GTK_WIDGET_VISIBLE (sbow->sbw.gui_obj.window))
    gtk_widget_show_all (sbow->sbw.gui_obj.window);
  else
  {
      sbow->clist_rows = 0;
      gtk_widget_destroy (sbow->sbw.gui_obj.window);
    }
}
