#include <stdio.h>
#include <stdlib.h>
#include <sys/errno.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <string.h>

#include <iostream.h>

#include "gtksheet/gtkcombobox.h"
#include "gtksheet/gtkbordercombo.h"
#include "gtksheet/gtkcolorcombo.h"
#include "gtksheet/gtksheet.h"
#include "gtksheet/gtksheetentry.h"

#include "../src/gpsim_def.h"
#include "../src/pic-processor.h"

#include "gui.h"

StatusBar_Window *status_bar_window;

void enter_callback(GtkWidget *widget, GtkWidget *entry);
void StatusBar_update_xref(cross_reference_to_gui *xref, int new_value);

/*
 * create_labeled_entry
 */

labeled_entry *create_labeled_entry(GtkWidget *box,char *label, int string_width)
{

  labeled_entry *le;


  le = (labeled_entry *)malloc(sizeof(labeled_entry));

  le->label = gtk_label_new (label);

  gtk_misc_set_alignment (GTK_MISC (le->label), 1.0, 0.5);
  gtk_widget_set_usize (le->label, 0, 15);
  gtk_box_pack_start (GTK_BOX (box), le->label, FALSE, FALSE, 0);
  gtk_widget_show (le->label);

  le->entry = gtk_entry_new ();
  gtk_signal_connect(GTK_OBJECT(le->entry), "activate",
		     GTK_SIGNAL_FUNC(enter_callback),
		     le->entry);
  gtk_entry_set_text (GTK_ENTRY (le->entry), "----");

  le->value = 0;

  gtk_widget_set_usize (le->entry,
			string_width * gdk_string_width (le->entry->style->font, "9") + 6,
			-1);
  gtk_box_pack_start (GTK_BOX (box), le->entry, FALSE, FALSE, 0);
  gtk_widget_show (le->entry);

  return(le);
}
/*
 * CreateStatusBar
 *
 * Create the status bar at the bottom of the window
 * 
 *
 * vbox_main - The box to which we will append.
 *
 */ 

void StatusBar_create(GtkWidget *vbox_main, StatusBar_Window *sbw)
{
  GtkWidget *hbox;

  /* --- Create h-box for holding the status line --- */
  hbox = gtk_hbox_new (FALSE, 0);

  /* --- Put up h-box --- */
  gtk_box_pack_start (GTK_BOX (vbox_main), hbox, FALSE, FALSE, 0);
  gtk_widget_show (hbox);

  sbw->status = create_labeled_entry(hbox,"Status: ", 4);
  sbw->status->parent = sbw;

  sbw->W = create_labeled_entry(hbox,"  W: ", 4);
  sbw->W->parent = sbw;

  sbw->pc = create_labeled_entry(hbox,"  PC: ", 6);
  sbw->pc->parent = sbw;

  sbw->cycles = create_labeled_entry(hbox,"  Cycles: ", 10);
  sbw->cycles->parent = sbw;

}


void StatusBar_update(StatusBar_Window *sbw)
{
  char buffer[20];

  pic_processor *pic;

  // get a manageable pointer 
  pic = sbw->gp->p;

  //update the displayed values

  sbw->status->value = pic->status.get_value();
  sprintf(buffer,"0x%02x",sbw->status->value);
  gtk_entry_set_text (GTK_ENTRY (sbw->status->entry), buffer);


  sbw->W->value = pic->W.get_value();
  sprintf(buffer,"0x%02x",sbw->W->value);
  gtk_entry_set_text (GTK_ENTRY (sbw->W->entry), buffer);

  sbw->pc->value = pic->pc.value;
  sprintf(buffer,"0x%04x",sbw->pc->value);
  gtk_entry_set_text (GTK_ENTRY (sbw->pc->entry), buffer);

  sbw->cycles->value = pic->cycles.value;
  sprintf(buffer,"0x%08x",sbw->cycles->value);
  gtk_entry_set_text (GTK_ENTRY (sbw->cycles->entry), buffer);

}

void StatusBar_update_xref(cross_reference_to_gui *xref, int new_value)
{
  StatusBar_Window *sbw;

  sbw  = (StatusBar_Window *) (xref->parent_window);
  StatusBar_update(sbw);
}

void StatusBar_new_processor(StatusBar_Window *sbw, GUI_Processor *gp)
{

  cross_reference_to_gui  *cross_reference;
  pic_processor *pic;

  if(sbw == NULL || gp == NULL)
    return;

  sbw->gp = gp;
  gp->status_bar = sbw;

  pic = sbw->gp->p;

  /* Now create a cross-reference link that the simulator can use to
   * send information back to the gui
   */

  cross_reference = (cross_reference_to_gui *) malloc(sizeof(cross_reference_to_gui));
  cross_reference->parent_window_type =   WT_status_bar;
  cross_reference->parent_window = (gpointer) sbw;
  cross_reference->data = (gpointer) sbw;
  cross_reference->update = StatusBar_update_xref;
  pic->pc.assign_xref((gpointer) cross_reference);


  StatusBar_update(sbw);

}
