/*
   Copyright (C) 1998 Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpasm; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>

#include "16bit-registers.h"
#include "16bit-processors.h"


#ifdef HAVE_GUI

void gui_update_object(gpointer gui_xref,int new_value);

#endif


//--------------------------------------------------
// member functions for the BSR class
//--------------------------------------------------
//
void  BSR::put(unsigned int new_value)
{
  value = new_value & 0x0f;
  trace.register_write(address,value);

  cpu->register_bank = &cpu->registers[ value << 8 ];


}

void  BSR::put_value(unsigned int new_value)
{

  put(new_value);

#ifdef HAVE_GUI

  gui_update_object(gui_xref,new_value);
  
  gui_update_object(cpu->indf.gui_xref,cpu->registers[new_value]->value);

#endif


}


//--------------------------------------------------
// member functions for the FSR class
//--------------------------------------------------
//
void  FSRL::put(unsigned int new_value)
{
  value = new_value & 0xff;
  trace.register_write(address,value);

  //  iam->fsr_delta = 0;
  iam->update_fsr_value();


}

void  FSRL::put_value(unsigned int new_value)
{

  put(new_value);

#ifdef HAVE_GUI

  gui_update_object(gui_xref,new_value);
  
  gui_update_object(cpu->indf.gui_xref,cpu->registers[new_value]->value);

#endif


}

void  FSRH::put(unsigned int new_value)
{
  value = new_value & 0x0f;
  trace.register_write(address,value);

  //  iam->fsr_delta = 0;
  iam->update_fsr_value();

}

void  FSRH::put_value(unsigned int new_value)
{

  put(new_value);

#ifdef HAVE_GUI

  gui_update_object(gui_xref,new_value);
  
  gui_update_object(cpu->indf.gui_xref,cpu->registers[new_value]->value);

#endif


}

void  INDF16::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  iam->fsr_value += iam->fsr_delta;
  iam->fsr_delta = 0;

  iam->put(new_value);

}

void INDF16::put_value(unsigned int new_value)
{
  put(new_value);
#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}

unsigned int INDF16::get(void)
{

  trace.register_read(address,value);

  iam->fsr_value += iam->fsr_delta;
  iam->fsr_delta = 0;

  return(iam->get());
}

unsigned int INDF16::get_value(void)
{
  return(iam->get_value());
}

//------------------------------------------------
// PREINC
unsigned int PREINC::get(void)
{

  trace.register_read(address,value);
  iam->preinc_fsr_value();

  return(iam->get());
}

unsigned int PREINC::get_value(void)
{
  return(iam->get_value());
}

void PREINC::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  iam->preinc_fsr_value();
  iam->put(new_value);
}

void PREINC::put_value(unsigned int new_value)
{
  put(new_value);

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}

//------------------------------------------------
// POSTINC
unsigned int POSTINC::get(void)
{

  trace.register_read(address,value);
  iam->postinc_fsr_value();

  return(iam->get());
}

unsigned int POSTINC::get_value(void)
{
  return(iam->get_value());
}

void POSTINC::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  iam->postinc_fsr_value();
  iam->put(new_value);
}

void POSTINC::put_value(unsigned int new_value)
{
  put(new_value);

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//------------------------------------------------
// POSTDEC
unsigned int POSTDEC::get(void)
{

  trace.register_read(address,value);
  iam->postdec_fsr_value();

  return(iam->get());
}

unsigned int POSTDEC::get_value(void)
{
  return(iam->get_value());
}

void POSTDEC::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  iam->postdec_fsr_value();
  iam->put(new_value);
}

void POSTDEC::put_value(unsigned int new_value)
{
  put(new_value);

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//------------------------------------------------
// PLUSW
unsigned int PLUSW::get(void)
{

  trace.register_read(address,value);

  int destination = iam->plusw_fsr_value();
  if(destination > 0)
    return(cpu->registers[destination]->get());
  else
    return 0;

}

unsigned int PLUSW::get_value(void)
{

  int destination = iam->plusw_fsr_value();
  if(destination > 0)
    return(cpu->registers[destination]->get_value());
  else
    return 0;

}

void PLUSW::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  int destination = iam->plusw_fsr_value();
  if(destination > 0)
    cpu->registers[destination]->put(new_value);
}

void PLUSW::put_value(unsigned int new_value)
{
  int destination = iam->plusw_fsr_value();
  if(destination > 0)
    cpu->registers[destination]->put_value(new_value);


#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
  if(destination > 0)
    gui_update_object(cpu->registers[destination]->gui_xref,new_value);
#endif

}

//------------------------------------------------

void Indirect_Addressing::init(_16bit_processor *new_cpu)
{

  fsrl.iam = this;
  fsrh.iam = this;
  indf.iam = this;
  preinc.iam = this;
  postinc.iam = this;
  postdec.iam = this;
  plusw.iam = this;

  cpu = new_cpu;

}

/*
 * put - Each of the indirect registers associated with this
 * indirect addressing class will call this routine to indirectly
 * write data.
 */
void Indirect_Addressing::put(unsigned int new_value)
{
  /*  unsigned int midbits;

  if( ((fsr_value & 0xfc7) == 0xfc3) || ((fsr_value & 0xfc4) == 0xfc4))
    {
      midbits = (fsr_value >> 3) & 0x7;
      if(midbits >= 3 && midbits <= 5)
	return;
    }
  */
  if(is_indirect_register(fsr_value))
    return;

  cpu->registers[fsr_value]->put(new_value);

}

/*
 * get - Each of the indirect registers associated with this
 * indirect addressing class will call this routine to indirectly
 * retrieve data.
 */
unsigned int Indirect_Addressing::get(void)
{
  //  unsigned int midbits;

  // See the comment in Indirect_Addressing::put about fsr address checking
  if(is_indirect_register(fsr_value))
    return 0;
  else
    /*
  if( ((fsr_value & 0xfc7) == 0xfc3) || ((fsr_value & 0xfc4) == 0xfc4))
    {
      midbits = (fsr_value >> 3) & 0x7;
      if(midbits >= 3 && midbits <= 5)
	return 0;
    }
    */

  return cpu->registers[fsr_value]->get();

}

/*
 * get - Each of the indirect registers associated with this
 * indirect addressing class will call this routine to indirectly
 * retrieve data.
 */
unsigned int Indirect_Addressing::get_value(void)
{
  /*
  unsigned int midbits;

  See the comment in Indirect_Addressing::put about fsr address checking

  if( ((fsr_value & 0xfc7) == 0xfc3) || ((fsr_value & 0xfc4) == 0xfc4))
    {
      midbits = (fsr_value >> 3) & 0x7;
      if(midbits >= 3 && midbits <= 5)
	return 0;
    }
  */
  if(is_indirect_register(fsr_value))
    return 0;
  else
    return cpu->registers[fsr_value]->get_value();

}

void Indirect_Addressing::put_fsr(unsigned int new_fsr)
{

  fsrl.put(new_fsr & 0xff);
  fsrh.put((new_fsr>>8) & 0x0f);

}


/*
 * update_fsr_value - This routine is called by the FSRL and FSRH
 * classes. It's purpose is to update the 16-bit (actually 12-bit)
 * address formed by the concatenation of FSRL and FSRH.
 *
 */

void Indirect_Addressing::update_fsr_value(void)
{

  if(current_cycle != cpu->cycles.value)
    {
      fsr_value = (fsrh.value << 8) |  fsrl.value;
      fsr_delta = 0;
    }
}

/*
 * preinc_fsr_value - This member function pre-increments the current
 * fsr_value. If the preinc access is a read-modify-write instruction
 * (e.g. bcf preinc0,1 ) then the increment operation should occur
 * only once. 
 */

void Indirect_Addressing::preinc_fsr_value(void)
{

  if(current_cycle != cpu->cycles.value)
    {
      fsr_value += (fsr_delta+1);
      fsr_delta = 0;
      current_cycle = cpu->cycles.value;
      put_fsr(fsr_value);
    }

}

void Indirect_Addressing::postinc_fsr_value(void)
{

  if(current_cycle != cpu->cycles.value)
    {
      fsr_value += fsr_delta;
      fsr_delta = 1;
      current_cycle = cpu->cycles.value;
      put_fsr(fsr_value+1);
      
    }
}

void Indirect_Addressing::postdec_fsr_value(void)
{

  if(current_cycle != cpu->cycles.value)
    {
      fsr_value += fsr_delta;
      fsr_delta = -1;
      current_cycle = cpu->cycles.value;
      put_fsr(fsr_value-1);
      
    }

}

int Indirect_Addressing::plusw_fsr_value(void)
{

  fsr_value += fsr_delta;
  fsr_delta = 0;

  unsigned int destination = (fsr_value + cpu->W.value) & _16BIT_REGISTER_MASK;
  if(is_indirect_register(destination))
    return -1;
  else
    return destination;

}

//------------------------------------------------
void Fast_Stack::init(_16bit_processor *new_cpu)
{
  cpu = new_cpu;
}

void Fast_Stack::push(void)
{
  w = cpu->W.value;
  status = cpu->status.value;
  bsr = cpu->bsr.value;

}

void Fast_Stack::pop(void)
{
  cout << "popping fast stack\n";
  cpu->W.put(w);
  cpu->status.put(status);
  cpu->bsr.put(bsr);

}

//--------------------------------------------------
// put_value - Change the program counter without affecting the cycle counter
//             (This is what's called if the user changes the pc.)

void Program_Counter16::put_value(unsigned int new_value)
{
  value = new_value & memory_size_mask;
  cpu->pcl.value = value & 0xff;
  cpu->pclath.value = (new_value >> 8) & 0xff;
  //  cpu16->pclatu.value = (new_value >> 16) & 0xff;

#ifdef HAVE_GUI

  gui_update_object(cpu->pcl.gui_xref,cpu->pcl.value);
  gui_update_object(cpu->pclath.gui_xref,cpu->pclath.value);
  //gui_update_object(cpu16->pclatu.gui_xref,cpu->pclatu.value);
  gui_update_object(gui_xref,new_value);

#endif
  

}

//------------------------------------------------
// TOSL
unsigned int TOSL::get(void)
{
  value = stack->get_tos() & 0xff;
  trace.register_read(address,value);
  return(value);

}

unsigned int TOSL::get_value(void)
{

  value = stack->get_tos() & 0xff;
  return(value);

}

void TOSL::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  stack->put_tos( stack->get_tos() & 0xffffff00 | new_value & 0xff);
}

void TOSL::put_value(unsigned int new_value)
{

  stack->put_tos( stack->get_tos() & 0xffffff00 | new_value & 0xff);

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//------------------------------------------------
// TOSH
unsigned int TOSH::get(void)
{
  value = (stack->get_tos() >> 8) & 0xff;
  trace.register_read(address,value);
  return(value);

}

unsigned int TOSH::get_value(void)
{

  value = (stack->get_tos() >> 8) & 0xff;
  return(value);

}

void TOSH::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  stack->put_tos( stack->get_tos() & 0xffff00ff | ( (new_value & 0xff) << 8));
}

void TOSH::put_value(unsigned int new_value)
{

  stack->put_tos( stack->get_tos() & 0xffff00ff | ( (new_value & 0xff) << 8));

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//------------------------------------------------
// TOSU
unsigned int TOSU::get(void)
{
  value = (stack->get_tos() >> 16) & 0x1f;
  trace.register_read(address,value);
  return(value);

}

unsigned int TOSU::get_value(void)
{

  value = (stack->get_tos() >> 16) & 0x1f;
  return(value);

}

void TOSU::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  stack->put_tos( stack->get_tos() & 0xffe0ffff | ( (new_value & 0x1f) << 16));

}

void TOSU::put_value(unsigned int new_value)
{

  stack->put_tos( stack->get_tos() & 0xffe0ffff | ( (new_value & 0x1f) << 16));

#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//------------------------------------------------
// STKPTR
unsigned int STKPTR::get(void)
{

  trace.register_read(address,value);
  return(value);

}

unsigned int STKPTR::get_value(void)
{

  return(value);

}

void STKPTR::put(unsigned int new_value)
{
  trace.register_write(address,new_value);

  value = new_value;
}

void STKPTR::put_value(unsigned int new_value)
{

  value = new_value;
#ifdef HAVE_GUI
  gui_update_object(gui_xref,new_value);
#endif

}


//--------------------------------------------------
//
Stack16::Stack16(void)
{
  stkptr.stack = this;
  tosl.stack = this;
  tosh.stack = this;
  tosu.stack = this;

}

void Stack16::push(unsigned int address)
{
  contents[stkptr.value & Stack16_MASK] = address;

  stkptr.value++;

  if((stkptr.value & Stack16_MASK) == 0)
    {
      // check the STVREN bit
      // if(STVREN) {reset(stack_over_flow); return;}
      stkptr.value |= 0x80;
    }

  stkptr.value &= 0xdf;

}

unsigned int Stack16::pop(void)
{

  if(stkptr.value & Stack16_MASK)
    return(contents[ (--stkptr.value) & Stack16_MASK]);
  else
    {
      // check the STVREN bit
      // if(STVREN) {reset(stack_over_flow); return;}
      stkptr.value |= 0x5f;
      return(contents[0]);
    }
}

void Stack16::reset(void)
{
  stkptr.value = 0;
}

unsigned int Stack16::get_tos(void)
{

  return (contents[stkptr.value & Stack16_MASK]);

}

void Stack16::put_tos(unsigned int new_tos)
{

  contents[stkptr.value & Stack16_MASK] = new_tos;

}
